import numpy as np
from gym import spaces

class Frozen_Lake_config:
    
    def __init__(self, direction_p):
        '''
        A configuration base class for the frozen lake
        direction_p: the slippery parameter (1/3 <= p <= 1)
        
        The map: ["SFFF", 
                  "FHFH", 
                  "FFFH", 
                  "HFFG"]
        '''
        self.direction_p = direction_p
        
        self.graph = {
        0: [-1, 4, -1, 1],
        1: [-1, 5, 0, 2],
        2: [-1, 6, 1, 3],
        3: [-1, 7, 2, -1 ],
        4: [0, 8, -1, 5],
        5: [-1, 4, -1, 1], # hole: return 0
        6: [2, 10, 5, 7],
        7: [-1, 4, -1, 1], #   hole: return 0
        8: [4, 12, -1, 9],
        9: [5, 13, 8, 10],
        10: [6, 14, 9, 11],
        11: [-1, 4, -1, 1], # hole: return 0
        12: [-1, 4, -1, 1], #  hole: return 0
        13: [9, -1, 12, 14],
        14: [10, -1, 13, 15],
        15: [0, 4, 0, 1] # goal: earn reward and return 0
        }
        
    
    @staticmethod
    def get_rnd_actions(action):
        if action == 0: 
            return [2,0,3]
        elif action == 1:
            return [2,1,3]
        elif action == 2:
            return [0,2,1]
        elif action == 3:
            return [0,3,1]
        else:
            raise
        
        
    def generate_trans_matrix(self, state_space, action_space):
        ret = np.zeros(shape = (len(action_space), len(state_space), len(state_space)))
        for a in action_space:
            for s in state_space:
                for ns in state_space:
                    a_ls = self.get_rnd_actions(a)
                    for ra in a_ls:
                        p = self.direction_p if ra == a else (1-self.direction_p)/2
                        if self.graph[s][ra] < 0:
                            ret[a, s, ns] += p if s == ns else 0.
                        else:
                            ret[a, s, ns] += p if ns == self.graph[s][ra] else 0.
        return ret

class Frozen_Lake_Env(Frozen_Lake_config):
    def __init__(self, random_factor):
        slippery_param = 1-2*random_factor
        Frozen_Lake_config.__init__(self, slippery_param)
        
        self.obs = 0
        self.act_space = self.control_space = tuple([0,1,2,3]) # u,d,l,r
        self.state_space = tuple(list(range(16)))
        # self.state_dim = 16
        # self.action_dim = 1
        
        self.obs_dim = 16
        self.act_dim = 1
        self.act_num = 4
        
        self.action_space = spaces.Discrete(4)
        self.observation_space = spaces.Box(0.0, 1.0, shape=(16,))

        self.trans_matrix = self.generate_trans_matrix(self.state_space, self.control_space)
        self.cnt = 0
        
        
    def step(self, action, current_obs = None):
        assert action in self.control_space, f"illegal actoin {action}"
        obs = self.obs if current_obs is None else current_obs
        self.obs = self._next_state(obs, action)
        r = 1 if self.obs == 15 else 0 # if the next state is the goal, give a positive reward.
        self.cnt += 1
        done = self.cnt >= 200
        return self.cont_obs, r, done, {}
    
    @property
    def cont_obs(self):
        tmp = np.zeros(shape = (16,))
        tmp[self.obs] = 1
        return tmp

    def reset(self, state="random"):
        self.cnt = 0
        if state == "random" or state is None:
            self.obs = np.random.choice(self.state_space)
        elif state in self.state_space:
            self.obs = state
        else:
            print(state)
            raise
        return self.cont_obs
    
    def _next_state(self, obs, action):
        next_s = np.random.choice(self.state_space, p = self.trans_matrix[action, obs, :])
        return next_s
    
    
    
if __name__ == "__main__":
    env = Frozen_Lake_Env(1.0)
    print(env.trans_matrix)
    print(env.reset(False))
    while True:
        ipt = input("Input directions (W, A, S, D) or reset (R): ")
        
        if ipt == 'r':
            print(env.reset(False))
        elif ipt == 'w': # up
            print(env.step(0) )
        elif ipt == 'a': # left
            print(env.step(2))
        elif ipt == 's': # down
            print(env.step(1))
        elif ipt == 'd': # right
            print(env.step(3))
        else:
            print(env.obs)
        
        
    