import gym
from gym import error, spaces, utils
from gym.utils import seeding
import numpy as np
import copy
from warnings import warn
from itertools import product

class PuddleEnv(gym.Env):
    metadata = {'render.modes': ['human', 'rgb_array']}

    def __init__(self, random_factor = 0.025, *, start=[0.2, 0.4], goal=[1.0, 1.0], goal_threshold=0.1, 
                 thrust=0.05, puddle_center=[[.3, .6], [.4, .5], [.8, .9]],
            puddle_width=[[.1, .03], [.03, .1], [.03, .1]]):
        self.start = np.array(start)
        self.goal = np.array(goal)
        self.goal_threshold = goal_threshold
        self.noise = random_factor
        self.random_factor = random_factor
        self.thrust = thrust
        self.puddle_center = [np.array(center) for center in puddle_center]
        self.puddle_width = [np.array(width) for width in puddle_width]

        self.action_space = spaces.Discrete(5)
        self.observation_space = spaces.Box(0.0, 1.0, shape=(2,))
        self.act_space = list(range(5)) # stay + 4 directions using index, 0:l, 1:r, 2:d, 3:r
        self.actions_true_space = [np.zeros(2) for i in range(5)] # four directions.
        for i in range(4):
            self.actions_true_space[i][i//2] = thrust * (i%2 * 2 - 1)
        print(self.actions_true_space) 
        self._seed()
        self.viewer = None
        self.cnt = 0
        self.done_flag = False
        
        self.start_pts = list(product(np.arange(0.1, 1.1, 0.1), np.arange(0.1, 1.1, 0.1)))
        
        self.obs_dim = 2
        self.act_dim = 1
        self.act_num = 5


    def _seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        if self.done_flag == True:
            warn(f'done_flag: {self.done_flag}')
            
        self.cnt += 1
        assert action in self.action_space,  "%r (%s) invalid"%(action, type(action))

        self.pos += self.actions_true_space[action] + self.np_random.uniform(low=-self.noise, high=self.noise, size=(2,)) # randomness is applied on both directions and magenitude.
        self.pos = np.clip(self.pos, 0.0, 1.0)

        reward = self._get_reward(self.pos)

        done = np.linalg.norm((self.pos - self.goal), ord=1) < self.goal_threshold 
        if done:
            self.done_flag = True
            reward = 0
        else:
            if self.cnt >= 200:
                # reward -= np.linalg.norm((self.pos - self.goal), ord=2)*10
                done = True
                self.done_flag = True


        return self.pos.copy(), reward, done, {}

    def _get_reward(self, pos):
        reward = -1.
        for cen, wid in zip(self.puddle_center, self.puddle_width):
            reward -= 3 * self._gaussian1d(pos[0], cen[0], wid[0]) * \
                self._gaussian1d(pos[1], cen[1], wid[1])

        return reward

    def _gaussian1d(self, p, mu, sig):
        return np.exp(-((p - mu)**2)/(2.*sig**2)) / (sig*np.sqrt(2.*np.pi))

    def reset(self, state = None):
        if self.start is None:
            raise NotImplementedError
            # self.pos = self.observation_space.sample()
            
        if state is None or state == "random":

            tmp_idx = np.random.choice(range(len(self.start_pts)))
            self.pos = self.start_pts[tmp_idx]          
        elif state == 'start':
            self.pos = copy.copy(self.start)
        else:
            self.pos = copy.copy(state)
            
        self.cnt = 0
        self.done_flag = False
        self.pos = np.array(self.pos)
        return self.pos.copy()

    # def _render(self, mode='human', close=False):
    #     if close:
    #         if self.viewer is not None:
    #             self.viewer.close()
    #             self.viewer = None
    #         return

    #     screen_width = 600
    #     screen_height = 400

    #     if self.viewer is None:
    #         from gym.envs.classic_control import rendering
    #         from gym_puddle.shapes.image import Image
    #         self.viewer = rendering.Viewer(screen_width, screen_height)

    #         import pyglet
    #         img_width = 100
    #         img_height = 100
    #         fformat = 'RGB'
    #         pixels = np.zeros((img_width, img_height, len(fformat)))
    #         for i in range(img_width):
    #             for j in range(img_height):
    #                 x = float(i)/img_width
    #                 y = float(j)/img_height
    #                 pixels[j,i,:] = self._get_reward(np.array([x,y]))

    #         pixels -= pixels.min()
    #         pixels *= 255./pixels.max()
    #         pixels = np.floor(pixels)

    #         img = pyglet.image.create(img_width, img_height)
    #         img.format = fformat
    #         data=[chr(int(pixel)) for pixel in pixels.flatten()]

    #         img.set_data(fformat, img_width * len(fformat), ''.join(data))
    #         bg_image = Image(img, screen_width, screen_height)
    #         bg_image.set_color(1.0,1.0,1.0)

    #         self.viewer.add_geom(bg_image)

    #         thickness = 5
    #         agent_polygon = rendering.FilledPolygon([(-thickness,-thickness),
    #          (-thickness,thickness), (thickness,thickness), (thickness,-thickness)])
    #         agent_polygon.set_color(0.0,1.0,0.0)
    #         self.agenttrans = rendering.Transform()
    #         agent_polygon.add_attr(self.agenttrans)
    #         self.viewer.add_geom(agent_polygon)

    #     self.agenttrans.set_translation(self.pos[0]*screen_width, self.pos[1]*screen_height)

    #     return self.viewer.render(return_rgb_array = mode=='rgb_array')
    

    
class True_Model:
    def __init__(self, random_factor):
        self.env = PuddleEnv(noise = random_factor)
        
    def plan(self, state, actions, num_trj):
        init_state = state.copy()
        ret = []
        for i in range(num_trj):
            trj, rewards = [init_state], []
            self.env.reset()
            self.env.pos = init_state
            done = False
            for act in actions:
                if not done:
                    state, rew, done, _ = self.env.step(act[0])
                else:
                    state, rew = state, rew # use previous ones.
                trj.append(state)
                rewards.append(rew)
                # if done:
                #     break
            # ret.append({'traj': np.array(trj), 'rewards': np.array(rewards)})
            ret.append(np.array(trj))
        return np.array(ret)
    
    
    
if __name__ == "__main__":
    import numpy as np
    env = PuddleEnv(start=[0.7, 0.7])
    print(env.action_space)
    obs = env.reset()
    ret = 0
    for k in range(205):
        act = np.random.choice([1, 3])#
        # act  = np.random.choice(env.action_space)
        next_obs, rew, done, _ = env.step(act)
        print(next_obs, rew, done)
        ret += rew
        if done:
            break
        
    print(ret, k)