import numpy as np
import gym
from gym import spaces
from gym.utils import seeding



class StormySwampyEnv(gym.Env):
    '''
    State Space: [0, 1]
    Action Space:  0: aggressively steering to the left,
                   1: mildly steering to the left, 
                   2: mildly steering to the right, 
                   3: aggressively steering to the right.
                   
    Reward Scheme: stay at safe zone ([0.2, 0.8]) : 1
                   first time step in danger zone : 0
                   stay at danger zone on the left : -5
                   stay at danger zone on the right : -8

    '''
    
    def __init__(self, random_factor = 0.05, *, 
                 field_length = 1, 
                 dead_zone = 0.2,
                 move_step = 0.4,
                 max_steps = 200,
                 ):
        self.random_factor = random_factor
        self.field_length = field_length
        self.dead_zone = dead_zone
        self.move_step = move_step
        self.max_steps = max_steps
        
        self.field_left_end = 0.
        self.field_right_end = field_length
        
        
        self.init_pos = lambda : np.random.uniform(0.4, 0.6)
        self.pos = np.array([self.init_pos()])
        self.counter = 0
        self.done = None
        
        self.random_gen = lambda : np.random.normal(0, random_factor)
        
        self.name = 'MagCar'
        
        self.INF_HOR = False
        
        self.act_dim = 1
        self.act_space = (0, 1, 2, 3)
        self.act_num = len(self.act_space)
        self.true_action_space = {0: -0.4, 
                                  1: -0.2, 
                                  2: 0.2,
                                  3: 0.4
                                  }
        self.obs_dim = 1
        
        
        self.action_space = spaces.Discrete(self.act_num)
        self.observation_space = spaces.Box(np.array([self.field_left_end]), 
                                            np.array([self.field_right_end]), dtype=np.float32)

        
        
    def at_end(self, loc):
        if loc >= self.field_right_end - self.dead_zone:
            return 1
        elif loc <= self.field_left_end + self.dead_zone:
            return -1
        else:
            return 0
        
        
    def step(self, action):
        # move = -self.move_step if action == 0 else self.move_step
        move = self.true_action_space[action]
        move = move+self.random_gen()
        prev_is_end = self.at_end(self.pos[0])
        
        # if (is_end == 1 and action == 1):  
        #     rew = -8
        # elif (is_end == -1 and action == 0):
        #     rew = -5
            
        # else:
        #     rew = 0

        self.pos = self.pos + move
        
        is_end = self.at_end(self.pos[0])
        if prev_is_end == -1:
            if is_end == -1:
                rew = -5
            else:
                rew = 0
        elif prev_is_end == 1:
            if is_end == 1:
                rew = -8
            else: 
                rew = 0
        elif prev_is_end == 0:
            if is_end == 0:
                rew = 0
            else:
                rew = 0
                
        
        
        if is_end == 1:
            self.pos = np.array([self.field_right_end])
            rew += 0
        elif is_end == -1:
            self.pos = np.array([self.field_left_end])
            rew += 0
        else:
            rew += 1
            
        self.counter += 1
        if self.done == True: print('reset should be called.')
        if self.counter > self.max_steps:
            self.done = True
        else:
            self.done = False
        return self.pos.copy(), rew, self.done, {}
    
    
    def reset(self, pos = None):
        if pos is None:
            self.pos = np.array([self.init_pos()])
        else:
            self.pos = np.array(pos)
        self.counter = 0
        self.done = False
        return self.pos.copy()
    
    
    
class True_Model:
    def __init__(self, random_factor):
        self.env = StormySwampyEnv(random_factor)
        
    def plan(self, state, actions, num_trj):
        init_state = state.copy()
        ret = []
        for i in range(num_trj):
            trj, rewards = [init_state], []
            self.env.reset(init_state)
            done = False
            for act in actions:
                if not done:
                    state, rew, done, _ = self.env.step(act)
                else:
                    state, rew = state, rew # use previous ones.
                trj.append(state)
                rewards.append(rew)
                # if done:
                #     break
            # ret.append({'traj': np.array(trj), 'rewards': np.array(rewards)})
            ret.append(np.array(trj))
        return np.array(ret)
        
        
if __name__ == '__main__':
    env = StormySwampyEnv(0.)
    obs = env.reset()
    for _ in range(205):
        act = np.random.choice(env.act_space)
        next_obs, rew, done, _ = env.step(act)
        print(obs, act, next_obs, rew)
        obs = next_obs
        
        