import tensorflow as tf
import numpy as np
from copy import deepcopy
from itertools import product
from gym import spaces

def initialize_uninitialized(sess):
    global_vars          = tf.global_variables()
    is_not_initialized   = sess.run([tf.is_variable_initialized(var) for var in global_vars])
    not_initialized_vars = [v for (v, f) in zip(global_vars, is_not_initialized) if not f]

    # print [str(i.name) for i in not_initialized_vars] # only for testing
    if len(not_initialized_vars):
        sess.run(tf.variables_initializer(not_initialized_vars))
        
        


class delayed_wrapper:
    def __init__(self, env, delay_steps, state_dim, action_dim, action_buffer = None, rendering = False):
        self.env = env
        self.observation_space = env.observation_space
        self.action_space = env.action_space
        self.delay_steps = delay_steps
        self.obs_dim = state_dim
        self.act_dim = action_dim
        assert len(action_buffer) == delay_steps, f"{len(action_buffer)} != {delay_steps}"
        self.action_buffer_default = action_buffer if action_buffer is not None else [0]*delay_steps
        self.render = rendering
        self.action_buffer = None
        self.reset()

    def step(self, action):
        app_action = self.action_buffer.pop(0)
        self.action_buffer.append(action)
        if self.render: self.env.render()
        return  self.env.step(app_action)

    def reset(self):
        self.action_buffer = deepcopy(self.action_buffer_default)
        return self.env.reset()
    
class delayed_wrapper_amdp:
    def __init__(self, env, delay_steps, state_dim, action_dim, action_buffer = None, rendering = False):
        self.env = env
        self._act_space = env.act_space
        _high = np.append(env.observation_space.high, [max(self._act_space)]*delay_steps)
        _low = np.append(env.observation_space.low, [min(self._act_space)]*delay_steps)
        self.observation_space = spaces.Box(low = _low, high = _high, dtype=env.observation_space.dtype)
        self.action_space = env.action_space
        self.delay_steps = delay_steps
        self.obs_dim = state_dim + delay_steps
        self.act_dim = action_dim
        assert len(action_buffer) == delay_steps, f"{len(action_buffer)} != {delay_steps}"
        self.action_buffer_default = action_buffer if action_buffer is not None else [0]*delay_steps
        self.render = rendering
        self.action_buffer = None
        self.reset()

    def step(self, action):
        app_action = self.action_buffer.pop(0)
        self.action_buffer.append(action)
        if self.render: self.env.render()
        obs, rew, done, d = self.env.step(app_action)
        return np.append(obs, self.action_buffer), rew, done, d

    def reset(self):
        self.action_buffer = deepcopy(self.action_buffer_default)
        return np.append(self.env.reset(), self.action_buffer)

def convert_ts(raw_data, steps):
    batchsize = len(raw_data)
    return np.array([raw_data[i:i+steps, :] for i in range(batchsize-steps)])

class Grid:
    def __init__(self, state_space_limits, num_bins):
        '''
        e.g., state_space_limits = [[-7, 7], [-0.5, 0.5]]. For a task whose state space is a
        product of [-7, 7]x[-0.5, 0.5]
        num_bins = [20, 10]
        '''
        self.state_space_limits, self.num_bins = state_space_limits, num_bins
        assert len(state_space_limits) == len(num_bins), (len(state_space_limits), len(num_bins))
        self.num_dim = len(num_bins)
        self.bins = []
        self.bin_widths = []
        self.bins_mid = []

        for n in range(self.num_dim):
            l, h = state_space_limits[n]
            num = num_bins[n] # num of seperator + 1 = num of bins
            self.bins.append(np.linspace(l, h, num = num-1))
            step = (h-l)/(num-2)
            self.bin_widths.append(step)
            self.bins_mid.append(np.linspace(l-step/2, h+step/2, num = num))

        self.num_disc_state = np.prod(np.array(num_bins))
        tmp_index_ls = [range(n) for n in num_bins] 
        
        self._map = {key:(n, self._get_bin_state_by_vec_label(key)) for n, key in enumerate(product(*tmp_index_ls))}
        self._map_values = list(self._map.values())
        self.disc_state_space = np.array([s[-1] for s in self._map_values])
        self._map_keys = list(self._map.keys())

    def _get_bin_state_by_vec_label(self, vec_label):
        return np.array([self.bins_mid[dim][n] for dim, n in enumerate(vec_label)])

    def _get_bin_state_by_label(self, label):
        return np.array(self._map_values[label][-1])

    def get_state(self, label, is_vec = False):
        if is_vec:
            return self._get_bin_state_by_vec_label(label)
        else:
            return self._get_bin_state_by_label(label)

    def get_state_from_dist(self, label_dist, num = None):
        idxs = np.random.choice(range(self.num_disc_state), p = label_dist, size = num)
        return self.disc_state_space[idxs]

    def search(self, state):
        vec_label = [np.digitize(ele, self.bins[n]) for n,ele in enumerate(state)]
        return self._map[tuple(vec_label)][0]

    def vec2label(self, vec_label):
        return self._map[tuple(vec_label)][0]

    def label2vec(self, label):
        return self._map_keys[label]

def add_env_model(cls):
    '''
    this function return a wrapped class
    '''
    class new_cls(cls):
        def __init__(self , *args, model, **kwargs):
            cls.__init__(self, *args, **kwargs)
            self.model = model
            
        def step(self, action):
            s = self.state
            a = action
            self.model.add_data(s, a)
            return cls.step(self, action)
        
    return new_cls
            



    